from django.contrib.auth.backends import BaseBackend
from rest_framework.exceptions import PermissionDenied

from .logto_init import start_logto_client
from .models import LogtoUser
from .utils import get_authenticated_logto_user, is_user_admin, get_or_create_admin


class LogtoBackend(BaseBackend):

    def authenticate(self, request, **kwargs):
        client = start_logto_client(request)
        logto_user: LogtoUser = get_authenticated_logto_user(client)
        user_is_admin = is_user_admin(logto_user)

        if user_is_admin:
            return get_or_create_admin(logto_user)
        else:
            raise PermissionDenied("You don't have the permissions to access Admin console")

    def get_user(self, user_id):
        logto_admin = LogtoUser.objects.filter(id=user_id).first()
        return logto_admin

    def has_perm(self, user_obj, perm, obj=None):
        return user_obj.is_admin and user_obj.is_authenticated



import os

from django.http import HttpResponseRedirect, HttpResponse
from django.views import View

from .logto_init import start_logto_client


class SigninView(View):
    async def get(self, request):
        uri = os.environ.get("LOGTO_API_REDIRECT_URI")
        client = start_logto_client(request)
        url = await client.signIn(redirectUri=uri)
        return HttpResponseRedirect(redirect_to=url)


class CallbackView(View):
    async def get(self, request):
        absolute_uri = request.build_absolute_uri()
        client = start_logto_client(request)
        try:
            await client.handleSignInCallback(absolute_uri)
            return HttpResponseRedirect(
                redirect_to=os.environ.get("LOGTO_CALLBACK_URI")
            )
        except Exception as e:
            return HttpResponse("Error: " + str(e))



import base64

import requests
from asgiref.sync import async_to_sync
from django.conf import settings
from django.http import HttpRequest
from .django_logto_client import DjangoLogtoClient
from .models import LogtoUser
from jose import jwt
from logto.models.response import UserInfoResponse
from rest_framework.exceptions import AuthenticationFailed, APIException


def get_authorization_header(request: HttpRequest) -> str:
    auth = request.headers.get("Authorization", '')

    return auth


def decode_token(token: str) -> dict:
    jwks_uri = requests.get(settings.LOGTO_JWKS_URIS)
    issuer = f"{settings.LOGTO_ENDPOINT}/oidc"
    jwks = jwks_uri.json()

    try:
        payload = jwt.decode(
            token,
            jwks,
            # The jwt encode algorithm retrieved along with jwks. ES384 by default
            algorithms=jwt.get_unverified_header(token).get('alg'),
            # The API's registered resource indicator in Logto
            audience=settings.LOGTO_AUDIENCE,
            issuer=issuer,
            options={
                'verify_at_hash': False
            }
        )
    except Exception as e:
        # exception handler
        raise AuthenticationFailed("Invalid token")

    return payload


def get_or_create_admin(logto_user: LogtoUser) -> LogtoUser:
    logto_admin = LogtoUser.objects.filter(sub=logto_user.sub).first()
    if logto_admin is None:
        logto_admin = LogtoUser.objects.create(
            sub=logto_user.sub,
            email=logto_user.email,
            username=logto_user.email
        )
        logto_admin.is_staff = True
        logto_admin.is_active = True
        logto_admin.is_superuser = True
        logto_admin.save()

    return logto_admin


def get_management_api_token():
    app_id = settings.LOGTO_ADMIN_CLIENT_ID
    app_secret = settings.LOGTO_ADMIN_CLIENT_SECRET
    encoded_credentials = base64.b64encode(
        f"{app_id}:{app_secret}".encode()
    ).decode()

    url = f"{settings.LOGTO_ENDPOINT}/oidc/token"
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': f"Basic {encoded_credentials}"
    }
    data = {
        'grant_type': 'client_credentials',
        'resource': 'https://default.logto.app/api',
        'scope': 'all'
    }

    response = requests.post(
        url,
        data=data,
        headers=headers
    )

    if response.status_code != 200:
        raise APIException(
            f"DjangoLogto exception on M2M interaction: {response.json()}"
        )
    access_token = response.json()["access_token"]

    return access_token


def get_authenticated_logto_user(client) -> LogtoUser:
    try:
        user_infos: UserInfoResponse = get_user_infos(client)

        access_token = get_management_api_token()
        management_api_url = f"{settings.LOGTO_ENDPOINT}/api/users/{user_infos.sub}"
        management_api_headers = {
            'Authorization': f"Bearer {access_token}"
        }
        user = requests.get(
            url=management_api_url,
            headers=management_api_headers
        ).json()

        user = LogtoUser(
            sub=user['id'],
            email=user['primaryEmail'],
            username=user['primaryEmail']
        )
        return user
    except Exception:
        raise AuthenticationFailed("You must pass by /auth/signin so signup")


def get_authenticated_logto_user_roles(user_infos: LogtoUser) -> list[dict]:
    access_token = get_management_api_token()
    management_api_url = f"{settings.LOGTO_ENDPOINT}/api/users/{user_infos.sub}/roles"
    management_api_headers = {
        'Authorization': f"Bearer {access_token}"
    }

    user_roles = requests.get(
        url=management_api_url,
        headers=management_api_headers
    )

    if user_roles.status_code != 200:
        raise APIException(
            f"DjangoLogto exception on ManagementAPI interaction: {user_roles.json()}"
        )

    return user_roles.json()


def is_user_admin(user_infos: LogtoUser) -> bool:
    user_is_admin = False
    user_roles = get_authenticated_logto_user_roles(user_infos)
    admin_tags = settings.LOGTO_ADMIN_TAGS if settings.LOGTO_ADMIN_TAGS is not None else ["admin"]

    for role in user_roles:
        if role['name'] in admin_tags:
            user_is_admin = True

    return user_is_admin


@async_to_sync
async def get_user_infos(client: DjangoLogtoClient):
    return await client.fetchUserInfo()



from setuptools import setup, find_packages
with open('README.md', 'r') as file:
    long_description = file.read()

setup(
    name='django-logto',
    version='1.0.4',
    packages=find_packages(),
    install_requires=[
        "django>=4.0",
        "djangorestframework>=3.0",
        "python-jose",
        "asgiref",
        "urllib3",
        "logto",
        "requests"
    ],
    author="Pulsation Studio",
    author_email="liam@pulsation.eco",
    description="DRF and Django Logto integration",
    url="https://github.com/pulsation-studio/django-logto",
    long_description=long_description,
    long_description_content_type="text/markdown"
)



from django.http import HttpRequest
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed

from .django_logto_client import DjangoLogtoClient
from .logto_init import start_logto_client
from .models import LogtoUser
from .utils import get_authenticated_logto_user, get_authorization_header, decode_token


class LogtoAuthentication(BaseAuthentication):
    def authenticate(self, request: HttpRequest):
        if request.accepted_renderer.format == 'api':
            client: DjangoLogtoClient = start_logto_client(request)
            authenticated_user = get_authenticated_logto_user(client)
            return authenticated_user, None

        auth: list[str] = get_authorization_header(request).split()

        if not auth or auth[0].lower() != "bearer":
            return None

        if len(auth) < 2:
            raise AuthenticationFailed("Invalid token format")

        elif auth[0] != 'Bearer':
            raise AuthenticationFailed("Token type not supported")

        decoded_token: dict = decode_token(auth[1])

        user_sub: str = decoded_token.get('sub')

        user = LogtoUser(
            sub=user_sub,
            username=None
        )

        return user, None



from django.urls import path

from .views import SigninView, CallbackView

urlpatterns = [
    path("signin/", SigninView.as_view(), name="signin"),
    path("callback/", CallbackView.as_view(), name="callback"),
]




from asgiref.sync import async_to_sync, sync_to_async
from django.conf import settings
from django.http import HttpRequest
from logto import Storage, LogtoConfig

from .django_logto_client import DjangoLogtoClient


class SessionStorage(Storage):

    def __init__(self, request: HttpRequest):
        self.request = request

    @sync_to_async
    def get(self, key: str):
        return self.request.session.get(key, '')

    @sync_to_async
    def set(self, key: str, value: str | None) -> None:
        self.request.session.__setitem__(key, value)

    @sync_to_async
    def delete(self, key: str) -> None:
        if key in self.request.session:
            self.request.session.__delitem__(key)


def start_logto_client(request: HttpRequest):
    client = DjangoLogtoClient(
        LogtoConfig(
            endpoint=f"{settings.LOGTO_ENDPOINT}/",
            appId=settings.LOGTO_API_CLIENT_ID,
            appSecret=settings.LOGTO_API_SECRET
        ),
        storage=SessionStorage(request)
    )
    return client


@async_to_sync
async def get_user_infos(client: DjangoLogtoClient):
    return await client.fetchUserInfo()



import time
import urllib
from typing import Optional

from asgiref.sync import sync_to_async
from logto import LogtoClient, LogtoConfig, Storage, LogtoException
from logto.LogtoClient import AccessTokenMap, AccessToken, SignInSession, InteractionMode
from logto.OidcCore import OidcCore
from logto.models.oidc import UserInfoScope, IdTokenClaims
from logto.models.response import TokenResponse, UserInfoResponse
from logto.utilities import removeFalsyKeys, OrganizationUrnPrefix


@sync_to_async
def validate_signin_session(signInSession):
    return SignInSession.model_validate_json(signInSession)


class DjangoLogtoClient(LogtoClient):
    def __init__(self, config: LogtoConfig, storage: Storage):
        super().__init__(config=config, storage=storage)

    async def _getAccessTokenMap(self) -> AccessTokenMap:
        """
        Get the access token map from storage.
        """
        accessTokenMap = await self._storage.get("accessTokenMap")
        try:
            return AccessTokenMap.model_validate_json(accessTokenMap)
        except:
            return AccessTokenMap(x={})

    async def _setAccessToken(self, resource: str, accessToken: str, expiresIn: int) -> None:
        """
        Set the access token for the given resource to storage.
        """
        accessTokenMap = await self._getAccessTokenMap()
        accessTokenMap.x[resource] = AccessToken(
            token=accessToken,
            expiresAt=int(time.time())
                      + expiresIn
                      - 60,  # 60 seconds earlier to avoid clock skew
        )
        await self._storage.set("accessTokenMap", accessTokenMap.model_dump_json())

    async def getAccessToken(self, resource: str = "") -> Optional[str]:
        """
        Get the access token for the given resource. If the access token is expired,
        it will be refreshed automatically. If no refresh token is found, None will
        be returned.
        """
        accessToken = await self._getAccessToken(resource)
        if accessToken is not None:
            return accessToken

        if (
                resource.startswith(OrganizationUrnPrefix)
                and UserInfoScope.organizations not in self.config.scopes
        ):
            raise LogtoException(
                "The `UserInfoScope.organizations` scope is required to fetch organization tokens"
            )

        refreshToken = await self._storage.get("refreshToken")
        if refreshToken is None:
            return None

        tokenResponse = await (await self.getOidcCore()).fetchTokenByRefreshToken(
            clientId=self.config.appId,
            clientSecret=self.config.appSecret,
            refreshToken=refreshToken,
            resource=resource,
        )

        await self._handleTokenResponse(resource, tokenResponse)
        return tokenResponse.access_token

    async def _getAccessToken(self, resource: str) -> Optional[str]:
        """
        Get the valid access token for the given resource from storage, no refresh will be
        performed.
        """
        accessTokenMap = await self._getAccessTokenMap()
        accessToken = accessTokenMap.x.get(resource, None)
        if accessToken is None or accessToken.expiresAt < int(time.time()):
            return None
        return accessToken.token

    async def _handleTokenResponse(
            self, resource: str, tokenResponse: TokenResponse
    ) -> None:
        """
        Handle the token response from the Logto server and store the tokens to storage.

        Resource can be an empty string, which means the access token is for UserInfo
        endpoint or the default resource.
        """
        if tokenResponse.id_token is not None:
            (await self.getOidcCore()).verifyIdToken(
                tokenResponse.id_token, self.config.appId
            )
            await self._storage.set("idToken", tokenResponse.id_token)

        if tokenResponse.refresh_token is not None:
            await self._storage.set("refreshToken", tokenResponse.refresh_token)

        await self._setAccessToken(
            resource, tokenResponse.access_token, tokenResponse.expires_in
        )

    async def _getSignInSession(self) -> Optional[SignInSession]:
        """
        Try to parse the current sign-in session from storage. If the value does not
        exist or parse failed, return None.
        """
        signInSession = await self._storage.get("signInSession")
        if signInSession is None:
            return None

        try:
            session = await validate_signin_session(signInSession)
            return session
        except:
            return None

    async def _setSignInSession(self, signInSession: SignInSession) -> None:
        await self._storage.set("signInSession", signInSession.model_dump_json())

    async def signIn(
            self, redirectUri: str, interactionMode: Optional[InteractionMode] = None
    ) -> str:
        """
        Returns the sign-in URL for the given redirect URI. You should redirect the user
        to the returned URL to sign in.

        By specifying the interaction mode, you can control whether the user will be
        prompted for sign-in or sign-up on the first screen. If the interaction mode is
        not specified, the default one will be used.

        Example:
          ```python
          return redirect(await client.signIn('https://example.com/callback'))
          ```
        """
        codeVerifier = OidcCore.generateCodeVerifier()
        codeChallenge = OidcCore.generateCodeChallenge(codeVerifier)
        state = OidcCore.generateState()
        signInUrl = await self._buildSignInUrl(
            redirectUri, codeChallenge, state, interactionMode
        )

        await self._setSignInSession(
            SignInSession(
                redirectUri=redirectUri,
                codeVerifier=codeVerifier,
                state=state,
            )
        )
        for key in ["idToken", "accessToken", "refreshToken"]:
            await self._storage.delete(key)

        return signInUrl

    async def signOut(self, postLogoutRedirectUri: Optional[str] = None) -> str:
        """
        Returns the sign-out URL for the given post-logout redirect URI. You should
        redirect the user to the returned URL to sign out.

        If the post-logout redirect URI is not provided, the Logto default post-logout
        redirect URI will be used.

        Note:
          If the OpenID Connect server does not support the end session endpoint
          (i.e. OpenID Connect RP-Initiated Logout), the function will throw an
          exception. Logto supports the end session endpoint.

        Example:
          ```python
          return redirect(await client.signOut('https://example.com'))
          ```
        """
        await self._storage.delete("idToken")
        await self._storage.delete("refreshToken")
        await self._storage.delete("accessTokenMap")
        endSessionEndpoint = (await self.getOidcCore()).metadata.end_session_endpoint

        if endSessionEndpoint is None:
            raise LogtoException(
                "End session endpoint not found in the provider metadata"
            )

        return (
                endSessionEndpoint
                + "?"
                + urllib.parse.urlencode(
            removeFalsyKeys(
                {
                    "client_id": self.config.appId,
                    "post_logout_redirect_uri": postLogoutRedirectUri,
                }
            )
        )
        )

    async def handleSignInCallback(self, callbackUri: str) -> None:
        """
        Handle the sign-in callback from the Logto server. This method should be called
        in the callback route handler of your application.
        """
        signInSession = await self._getSignInSession()

        if signInSession is None:
            raise LogtoException("Sign-in session not found")

        # Validate the callback URI without query matches the redirect URI
        parsedCallbackUri = urllib.parse.urlparse(callbackUri)
        if (
                parsedCallbackUri.path
                != urllib.parse.urlparse(signInSession.redirectUri).path
        ):
            raise LogtoException(
                "The URI path does not match the redirect URI in the sign-in session"
            )

        query = urllib.parse.parse_qs(parsedCallbackUri.query)
        if "error" in query:
            raise LogtoException(query["error"][0])

        if signInSession.state != query.get("state", [None])[0]:
            raise LogtoException("Invalid state in the callback URI")

        code = query.get("code", [None])[0]
        if code is None:
            raise LogtoException("Code not found in the callback URI")

        tokenResponse = await (await self.getOidcCore()).fetchTokenByCode(
            clientId=self.config.appId,
            clientSecret=self.config.appSecret,
            redirectUri=signInSession.redirectUri,
            code=code,
            codeVerifier=signInSession.codeVerifier,
        )

        await self._handleTokenResponse("", tokenResponse)
        await self._storage.delete("signInSession")

    async def getIdToken(self) -> Optional[str]:
        """
        Get the ID Token string. If you need to get the claims in the ID Token, use
        `getIdTokenClaims` instead.
        """
        return await self._storage.get("idToken")

    async def getIdTokenClaims(self) -> IdTokenClaims:
        """
        Get the claims in the ID Token. If the ID Token does not exist, an exception
        will be thrown.
        """
        idToken = await self._storage.get("idToken")
        if idToken is None:
            raise LogtoException("ID Token not found")

        return OidcCore.decodeIdToken(idToken)

    async def getRefreshToken(self) -> Optional[str]:
        """
        Get the refresh token string.
        """
        return await self._storage.get("refreshToken")

    async def isAuthenticated(self) -> bool:
        """
        Check if the user is authenticated by checking if the ID Token exists.
        """
        return await self._storage.get("idToken") is not None

    async def fetchUserInfo(self) -> UserInfoResponse:
        """
        Fetch the user information from the UserInfo endpoint. If the access token
        is expired, it will be refreshed automatically.
        """
        accessToken = await self.getAccessToken()
        return await (await self.getOidcCore()).fetchUserInfo(accessToken)
